# AUTOGENERATED! DO NOT EDIT! File to edit: ../00_file.ipynb.

# %% auto 0
__all__ = ['read_patterns']

# %% ../00_file.ipynb 3
from collections.abc import Iterable, Sequence
import re

# %% ../00_file.ipynb 5
def read_patterns(
    lines: Iterable[str] # e.g. open('hyph-fi.tex').readlines()
) -> tuple[Sequence[str], Sequence[str]]: # patterns and exceptions
    """Read a TeX hyphenation pattern file; not a complete parser but hopefully enough
    for hyph-*.tex"""
    comment = re.compile(r' *%.*')
    state = 0
    patterns = []
    exceptions = []
    for line in lines:
        line = line.strip()
        line = comment.sub('', line)
        nextstate = state
        if line == r'\patterns{':
            state = 1
            continue
        if line == r'\hyphenation{':
            state = 2
            continue
        if '}' in line:
            line = re.sub(r'}.*', '', line)
            nextstate = 0
        if not line:
            continue           
        if state == 1:
            patterns.extend(line.split())
        elif state == 2:
            exceptions.extend(line.split())
        state = nextstate
    return patterns, exceptions



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/shyster',
                'doc_host': 'https://jkseppan.github.io',
                'git_url': 'https://github.com/jkseppan/shyster',
                'lib_path': 'shyster'},
  'syms': { 'shyster.file': {'shyster.file.read_patterns': ('file.html#read_patterns', 'shyster/file.py')},
            'shyster.html': { 'shyster.html.analyze_soups': ('html.html#analyze_soups', 'shyster/html.py'),
                              'shyster.html.hyphenate_soup': ('html.html#hyphenate_soup', 'shyster/html.py')},
            'shyster.hyphenation': { 'shyster.hyphenation.add_hyphens': ('hyphenation.html#add_hyphens', 'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator': ('hyphenation.html#hyphenator', 'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.__call__': ( 'hyphenation.html#hyphenator.__call__',
                                                                                  'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.__init__': ( 'hyphenation.html#hyphenator.__init__',
                                                                                  'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.add_exception': ( 'hyphenation.html#hyphenator.add_exception',
                                                                                       'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.add_prefix_exception': ( 'hyphenation.html#hyphenator.add_prefix_exception',
                                                                                              'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.hyphenate': ( 'hyphenation.html#hyphenator.hyphenate',
                                                                                   'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.rm_exception': ( 'hyphenation.html#hyphenator.rm_exception',
                                                                                      'shyster/hyphenation.py'),
                                     'shyster.hyphenation.hyphenator.rm_prefix_exception': ( 'hyphenation.html#hyphenator.rm_prefix_exception',
                                                                                             'shyster/hyphenation.py')},
            'shyster.main': {'shyster.main.hyphenate_html_file': ('main.html#hyphenate_html_file', 'shyster/main.py')},
            'shyster.pattern': { 'shyster.pattern._cvt': ('pattern.html#_cvt', 'shyster/pattern.py'),
                                 'shyster.pattern.convert_exceptions': ('pattern.html#convert_exceptions', 'shyster/pattern.py'),
                                 'shyster.pattern.convert_patterns': ('pattern.html#convert_patterns', 'shyster/pattern.py')}}}


# AUTOGENERATED! DO NOT EDIT! File to edit: ../01_pattern.ipynb.

# %% auto 0
__all__ = ['convert_patterns', 'convert_exceptions']

# %% ../01_pattern.ipynb 3
import string, re
import itertools as it
from collections.abc import Iterable, Mapping
import datrie  # type: ignore

# %% ../01_pattern.ipynb 5
def _cvt(
    pattern: str  # pattern as read from the TeX patterns file
) -> tuple[int, ...]:  # position i has the weight of the slot before character i
    res = [0 for _ in pattern]
    pos = 0
    for ch in pattern:
        if ch in string.digits:
            res[pos] = int(ch)
        else:
            pos += 1
    return tuple(res[:pos+1])

# %% ../01_pattern.ipynb 8
def convert_patterns(
    patterns: Iterable[str]  # TeX style patterns
) -> datrie.Trie:  # trie mapping matched substrings to weights
    num = re.compile('[0-9]')
    patterns = sorted(patterns, key=lambda x: num.sub('', x))
    alphabet = set(it.chain.from_iterable(patterns)) - set(string.digits) | {'\x1F'}
    trie = datrie.Trie(alphabet)
    for pat in patterns:
        pat = pat.replace('.', '\x1f')
        trie[num.sub('', pat)] = _cvt(pat)
    return trie

# %% ../01_pattern.ipynb 12
def convert_exceptions(
    exceptions: Iterable[str]
) -> Mapping[str, tuple[str,...]]:  # mapping from word to word parts
    return {w.replace('-', ''): tuple(w.split('-')) for w in exceptions}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../02_hyphenation.ipynb.

# %% auto 0
__all__ = ['hyphenator']

# %% ../02_hyphenation.ipynb 4
import datrie  # type: ignore
import itertools as it
import string
from collections.abc import Sequence, Mapping, Iterable
from shyster.file import read_patterns
from shyster.pattern import convert_patterns, convert_exceptions
from pathlib import Path

# %% ../02_hyphenation.ipynb 6
def add_hyphens(
    s: str,  # word to hyphenate
    positions: Sequence[int],  # positions to insert hyphens (increasing order)
    hyphen: str='-'  # hyphen character
) -> str:  # word with hyphens
    i0, i1 = it.tee(iter(positions))
    i0 = it.chain((0,), i0)
    i1 = it.chain(i1, (len(s),))
    substrings = (s[p0:p1] for (p0,p1) in zip(i0, i1))
    return hyphen.join(substrings).strip(hyphen)

# %% ../02_hyphenation.ipynb 10
class hyphenator:
    """Hyphenates words"""
    __slots__ = ('trie', 'exceptions', 'prefix_exceptions', 'hyphen', 'lefthyphenmin', 'righthyphenmin', 'alphabet')
    trie: datrie.Trie  # the patterns from `convert_patterns`
    exceptions: dict[str, tuple[str,...]]  # return value from `pattern.convert_exceptions`
    prefix_exceptions: datrie.Trie  # prefix exceptions (not in TeX hyphenation patterns)
    hyphen: str  # hyphen character
    lefthyphenmin: int  # at least this many characters before the first hyphen
    righthyphenmin: int  # at least this many characters after the last hyphen
    alphabet: str  # what characters can appear in exceptions

    def __init__(
        self,
        initializer: (str   # filename of hyphen.tex, or an iterable of its lines, or None
                      | Path
                      | Iterable[str]
                      | None),
        hyphen: str='-',
        lefthyphenmin: int=2,
        righthyphenmin: int=3,
        alphabet: str|None=None,  # alphabet; None for ASCII default
    ):
        if initializer is None:
            # the user will set these up explicitly
            self.trie = datrie.Trie('')
            self.exceptions = {}
        else:
            f = None
            if isinstance(initializer, (str, Path)):
                f = open(initializer, 'rt')
                it: Iterable[str] = f.readlines()
            elif isinstance(initializer, Iterable):
                it = initializer
            else:
                raise TypeError(f"don't know how to use {type(initializer)}")
            pat, exc = read_patterns(it)
            if f:
                f.close()
            self.trie = convert_patterns(pat)
            self.exceptions = dict(convert_exceptions(exc))
            
        self.hyphen = hyphen
        self.lefthyphenmin = lefthyphenmin
        self.righthyphenmin = righthyphenmin
        if alphabet is None:
            alphabet = string.ascii_letters
        self.alphabet = alphabet + '\x1F'
        self.prefix_exceptions = datrie.Trie(self.alphabet)
    
    def __call__(self, word: str):
        return self.hyphenate(word)
    
    def add_exception(
        self,
        word: str,  # word to add, possibly with `-` characters to indicate hyphenation points
        split: tuple[str,...] | None=None,  # how to split the word, or None to split at `-` characters
    ):
        if split is None:
            self.exceptions.update(convert_exceptions([word]))
        else:
            self.exceptions[word] = split
    
    def add_prefix_exception(
        self,
        prefix: str,
        split: tuple[str,...] | None=None,
    ):
        if split is None:
            (key, value), = convert_exceptions([prefix]).items()
            self.prefix_exceptions[key] = value
        else:
            self.prefix_exceptions[prefix] = split
    
    def rm_exception(
        self,
        word: str  # word to make unexceptional, without hyphens
    ):
        del self.exceptions[word]
        
    def rm_prefix_exception(
        self,
        prefix: str  # prefix to make unexceptional, without hyphens
    ):
        if self.prefix_exceptions is not None:
            del self.prefix_exceptions[prefix]
        
    def hyphenate(self, word: str) -> str:
        if (result := self.exceptions.get(word)):
            return self.hyphen.join(result)
        if (item := self.prefix_exceptions.longest_prefix_item(word, default=None)) is not None:
            prefix, value = item
            return self.hyphen.join(value) + word[len(prefix):]
        word = f'\x1f{word}\x1f'
        weights = bytearray(len(word))
        for pos in range(len(word)-1):
            for rule in self.trie.iter_prefix_values(word[pos:]):
                for i, w in enumerate(rule):
                    weights[pos+i-1] = max(weights[pos+i-1], w)
        positions = [i for (i,w) in enumerate(weights)
                     if w&1==1 and i>=self.lefthyphenmin and i<=len(word)-2-self.righthyphenmin]
        return add_hyphens(word[1:-1], positions, hyphen=self.hyphen)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../04_main.ipynb.

# %% auto 0
__all__ = ['hyphenate_html_file']

# %% ../04_main.ipynb 3
from shyster.file import read_patterns
from shyster.pattern import convert_patterns, convert_exceptions
from shyster.hyphenation import hyphenator
from shyster.html import hyphenate_soup
import bs4

from collections.abc import Callable, Set
from typing import Type

# %% ../04_main.ipynb 5
def hyphenate_html_file(
    html_filename: str,  # to be read
    output_filename: str,  # to be written
    patterns_filename: str,  # e.g. hyphen.tex or hyph-fi.tex
    hyphen: str='\N{SOFT HYPHEN}',  # what hyphen character to use
    lefthyphenmin: int=2,  # at least this many characters before the first hyphen
    righthyphenmin: int=3,  # at least this many characters after the last hyphen
    exclude_classes: tuple[Type[bs4.element.PageElement],...]=()  # exclude these bs4 elements (empty to use default)
) -> None:
    """Read html_filename and output output_filename, inserting hyphens in words."""
    hyph = hyphenator(patterns_filename, hyphen=hyphen, lefthyphenmin=lefthyphenmin, righthyphenmin=righthyphenmin)
    with open(html_filename, 'rt') as f:
        soup = bs4.BeautifulSoup(f, 'lxml')
    kwargs = {'exclude_classes': exclude_classes} if exclude_classes else {}
    hyphenate_soup(soup, hyph, **kwargs)
    with open(output_filename, 'wt') as f:
        f.write(str(soup))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../03_html.ipynb.

# %% auto 0
__all__ = ['hyphenate_soup', 'analyze_soups']

# %% ../03_html.ipynb 4
from collections.abc import Callable, Set
from collections import Counter
from typing import Type
import bs4
import re

# %% ../03_html.ipynb 5
word_or_punct_re = re.compile(r'\w+|[^\w]+')
word_re = re.compile(r'\w+')
default_exclude_classes = (
    bs4.element.PreformattedString,
    bs4.element.Stylesheet,
    bs4.element.Script,
    bs4.element.RubyTextString, # type: ignore
    bs4.element.RubyParenthesisString, # type: ignore
)

# %% ../03_html.ipynb 7
def hyphenate_soup(
    soup: bs4.BeautifulSoup,  # soup to be modified
    hyphenator: Callable[[str], str],  # hyphenator
    exclude_classes: tuple[Type[bs4.element.PageElement],...]=default_exclude_classes,  # do not modify inside these
) -> None:
    """Call hyphenator on words that appear in suitable elements of soup,
    and replace the contents of those elements. Suitable elements are those
    containing text whose class is not (a subclass of something) in
    exclude_classes."""
    for t in soup.find_all(string=True):
        if isinstance(t, exclude_classes):
            continue
        pieces = re.findall(word_or_punct_re, str(t.string))
        for i, word in enumerate(pieces):
            if not word_re.fullmatch(word):
                continue
            if (title := word.istitle()) or word.islower():
                hyphenated = hyphenator(word)
                if title:
                    pieces[i] = hyphenated.title()
                else:
                    pieces[i] = hyphenated
        new = ''.join(pieces)
        t.string.replace_with(new)


# %% ../03_html.ipynb 10
def analyze_soups(
    dinner: list[bs4.BeautifulSoup],  # soups to be read
    exclude_classes: tuple[Type[bs4.element.PageElement],...]=default_exclude_classes,  # do not look inside these
) -> Counter:
    """Count words appearing in all soups."""
    counter: Counter = Counter()
    for soup in dinner:
        for t in soup.find_all(string=True):
            if isinstance(t, exclude_classes):
                continue
            pieces = re.findall(word_or_punct_re, str(t.string))
            counter.update(word.lower() for word in pieces
                           if word_re.fullmatch(word) and word.istitle() or word.islower())
    return counter
